from typing import List, Optional
from datetime import datetime as dt
from datetime import timedelta as td

# app
from . import models
from . import wrappers
from .utils import notnull


def find_user(user_id: Optional[int], email: Optional[str]) -> Optional[models.User]:
	"""
	Used for finding models.User object by either id or email.
	:param user_id: id of user being queried
	:param email: email of user being queried
	:return: models.User object
	"""

	if user_id is not None:
		return models.User.get_or_none(id=user_id)
	elif email is not None:
		return models.User.get_or_none(email=email)
	else:
		notnull(None)  # both user_id and email are None


def is_participant(user: models.User, campaign: models.Campaign) -> bool:
	"""
	Checks whether a user is a campaign's participant or not
	:param user: user being checked
	:param campaign: campaign being checked
	:return: true if user is campaign's participant, false if not
	"""

	return models.Participant.filter(campaign=notnull(campaign).id, user=notnull(user).id).exists()


def is_supervisor(user: models.User, campaign: models.Campaign) -> bool:
	"""
	Checks whether a user is a campaign's supervisor or not
	:param user: user being checked
	:param campaign: campaign being checked
	:return: true if user is campaign's supervisor, false if not
	"""

	return models.Supervisor.filter(campaign=campaign, user=user).exists()


def get_participant(user: models.User, campaign: models.Campaign) -> models.Participant:
	"""
	Returns a participant object depending on the user and campaign provided
	:param user: user key to search for a participant object
	:param campaign: campaign key to search for a participant object
	:return: participant object
	"""

	return models.Participant.get_or_none(campaign=notnull(campaign), user=notnull(user))


def get_supervisor(user: models.User, campaign: models.Campaign) -> models.Supervisor:
	"""
	Returns a supervisor object depending on the user and campaign provided
	:param user: user key to search for a supervisor object
	:param campaign: campaign key to search for a supervisor object
	:return: supervisor object
	"""

	return models.Supervisor.get_or_none(campaign=notnull(campaign), user=notnull(user))


def get_campaign_participants(campaign: models.Campaign) -> List[models.Participant]:
	"""
	Returns list of participants of a campaign
	:param campaign: campaign being queried
	:return: list of campaign's participants
	"""

	return models.Participant.filter(campaign=notnull(campaign))


def get_campaign_participants_count(campaign: models.Campaign) -> int:
	"""
	Returns count of participants of a campaign
	:param campaign: campaign being queried
	:return: number of campaign's participants
	"""

	return models.Participant.filter(campaign=notnull(campaign)).count()


def get_campaign_supervisors(campaign: models.Campaign) -> List[models.Supervisor]:
	"""
	Returns list of a campaign's supervisors
	:param campaign: campaign being queried
	:return: list of campaign's supervisors
	"""

	return models.Supervisor.filter(campaign=notnull(campaign))


def get_campaign(campaign_id: int) -> Optional[models.Campaign]:
	"""
	Used for finding models.Campaign object by id.
	:param campaign_id: id of campaign being queried
	:return: models.Campaign object
	"""

	return models.Campaign.get_or_none(id=notnull(campaign_id))


def get_supervisor_campaigns(user: models.User) -> List[models.Campaign]:
	"""
	Filter campaigns by supervisor (when researcher wants to see the list of their campaigns)
	:param user: the supervisor
	:return: list of supervisor's campaigns
	"""

	return list(map(lambda supervisor: supervisor.campaign, models.Supervisor.filter(user=notnull(user))))


def find_data_source(
	data_source_id: Optional[int],
	name: Optional[str],
) -> Optional[models.DataSource]:
	"""
	Used for finding DataSource object by either id or name.
	:param data_source_id: id of data source being queried
	:param name: name of data source being queried
	:return: DataSource object (if found)
	"""

	if data_source_id is not None:
		return models.DataSource.get_or_none(id=data_source_id)
	elif name is not None:
		return models.DataSource.get_or_none(name=name)
	else:
		notnull(None)  # both data_source_id and name are None


def get_all_data_sources() -> List[models.DataSource]:
	"""
	List of all data sources in database
	:return: the list of data sources
	"""

	return models.DataSource.select()


def get_campaign_data_sources(campaign: models.Campaign) -> List[models.DataSource]:
	"""
	Returns list of a campaign's data sources
	:param campaign: campaign being queried
	:return: list of campaign's data sources
	"""

	return list(map(lambda campaign_data_source: campaign_data_source.data_source, models.CampaignDataSources.filter(campaign=notnull(campaign))))


def get_next_k_data_records(participant: models.Participant, data_source: models.DataSource, from_ts: dt, k: int) -> List[wrappers.DataRecord]:
	"""
	Retrieves next k data records from database
	:param participant: participant that has refernece to user and campaign
	:param data_source: type of data to retrieve
	:param from_ts: starting timestamp
	:param k: max amount of records to query
	:return: list of data records
	"""

	return wrappers.DataTable.select_next_k(participant=notnull(participant), data_source=notnull(data_source), from_ts=notnull(from_ts), limit=notnull(k))


def get_filtered_data_records(participant: models.Participant, data_source: models.DataSource, from_ts: dt = None, till_ts: dt = None) -> List[wrappers.DataRecord]:
	"""
	Retrieves filtered data based on provided range (start and end timestamps)
	:param participant: participant that has refernece to user and campaign
	:param data_source: type of data to retrieve
	:param from_ts: starting timestamp
	:param till_ts: ending timestamp
	:return: list of data records
	"""

	return wrappers.DataTable.select_range(participant=notnull(participant), data_source=notnull(data_source), from_ts=notnull(from_ts), till_ts=notnull(till_ts))


def get_participants_latest_stats(participant: models.Participant) -> wrappers.ParticipantStats:
	"""
	returns latest available statistics of a participant
	:param participant: participant being queried
	:return: latest stats
	"""

	return wrappers.ParticipantStats(participant=notnull(participant))


def get_filtered_amount_of_data(participant: models.Participant, data_source: models.DataSource, from_ts: dt, till_ts: dt) -> int:
	"""
	Computes and returns the amount of data during specified period
	:param participant: participant being queried
	:param data_source: data source being queried
	:param from_ts: start of the search range
	:param till_ts: end of the search range
	:return: amount of samples in the range with specified filters
	"""

	current: models.HourlyStats = models.HourlyStats.get_or_none(participant=notnull(participant), data_source=notnull(data_source), ts=till_ts.replace(minute=0, second=0, microsecond=0) + td(hours=1))
	if not current:
		current: models.HourlyStats = models.HourlyStats.filter(
			participant=notnull(participant),
			data_source=notnull(data_source),
		).order_by(models.HourlyStats.ts.desc()).limit(1)

		if not current:
			return 0

	if data_source not in current:
		return 0

	back_then: models.HourlyStats = models.HourlyStats.filter(participant=notnull(participant), data_source=notnull(data_source), ts=from_ts.replace(minute=0, second=0, microsecond=0))
	if not back_then or data_source not in back_then:
		if data_source in current.amounts:
			return current.amounts[data_source]
		else:
			return 0

	return current.amounts[data_source] - back_then.amounts[data_source]


def is_campaign_data_source(campaign: models.Campaign, data_source: models.Campaign) -> None:
	"""
	Checks if data source is being used by a campaign
	:param campaign: the campaign being queried
	:param data_source: data source being queried
	:return: whether data source is used by campaign
	"""

	return models.CampaignDataSources.filter(campaign=campaign, data_source=data_source).exists()



from datetime import datetime as dt
from typing import Dict, List, Optional

# app
from . import models
from . import wrappers
from . import selectors
from .utils import notnull


def create_user(email: str, name: str, session_key: str) -> models.User:
	"""
	Creates a user object in database and returns User object
	:param email: email of new user
	:param name: name of new user
	:param session_key: session_key for the new user
	:return:
	"""

	return models.User.create(email=notnull(email), name=notnull(name), session_key=notnull(session_key))


def set_user_session_key(user: models.User, new_session_key: str) -> None:
	"""
	Updates a user's session key (that is used for authentication)
	:param user: the user
	:param new_session_key: new session key
	:return: None
	"""

	user.session_key = notnull(new_session_key)
	user.save()


def add_participant_to_campaign(add_user: models.User, campaign: models.Campaign) -> bool:
	"""
	Binds user with campaign, making a participant.
	After binding is done, creates a new Data table for storing the participant's data.
	:param add_user: User object to be bound to a campaign
	:param campaign: Campaign object that user binds with
	:return: whether user has been bound
	"""

	if selectors.is_participant(user=notnull(add_user), campaign=notnull(campaign)):
		return False

	# 1. bind the user to campaign
	participant = models.Participant.create(campaign=campaign, user=add_user)

	# 2. create a new data table for the participant
	wrappers.DataTable.create(participant=participant)

	return True


def add_supervisor_to_campaign(
	new_user: models.User,
	supervisor: models.Supervisor,
) -> bool:
	"""
	Binds user with campaign, making a supervisor.
	:param new_user: User object to be bound to a campaign as a supervisor
	:param supervisor: Supervisor object that has reference to campaign (initially the owner is the first supervisor)
	:return: whether user has been bound (as supervisor)
	"""

	campaign: models.Campaign = notnull(supervisor).campaign

	if selectors.is_supervisor(user=notnull(new_user), campaign=notnull(campaign)):
		return False

	models.Supervisor.create(campaign=campaign, user=new_user)
	return True


def remove_supervisor_from_campaign(oldSupervisor: models.Supervisor) -> None:
	"""
	Unbinds a (supervisor) user from campaign.
	:param oldSupervisor: supervisor representing the binding between a user and a campaign.
	:return: None
	"""

	campaign: models.Campaign = notnull(oldSupervisor).campaign
	if oldSupervisor.user != campaign.owner:
		oldSupervisor.delete()


def create_campaign(owner: models.User, name: str, start_ts: dt, end_ts: dt, data_sources: List[models.DataSource]) -> models.Campaign:
	"""
	Creates a campaign object in database and returns Campaign object
	:param owner: owner (User instance) of the new campaign
	:param name: title of the campaign
	:param start_ts: when campaign starts
	:param end_ts: when campaign ends
	:param data_sources: data sources of the campaign
	:return: newly created Campaign instance
	"""

	# 1. create a campaign
	campaign = models.Campaign.create(owner=notnull(owner), name=notnull(name), start_ts=start_ts, end_ts=end_ts)

	# 2. add owner as a supervisor
	models.Supervisor.create(campaign=campaign, user=owner)

	# 3. add campaign's data sources
	for data_source in data_sources:
		add_campaign_data_source(campaign=campaign, data_source=data_source)

	return campaign


def add_campaign_data_source(campaign: models.Campaign, data_source: models.Campaign) -> None:
	"""
	Adds the data source to campaign
	:param campaign: the campaign to add data source to
	:param data_source: data source being added
	:return: None
	"""

	if selectors.is_campaign_data_source(campaign=campaign, data_source=data_source):
		return

	models.CampaignDataSources.create(campaign=campaign, data_source=data_source)


def remove_campaign_data_source(campaign: models.Campaign, data_source: models.Campaign) -> None:
	"""
	Removes the data source from campaign
	:param campaign: the campaign to remove data source from
	:param data_source: data source being removed
	:return: None
	"""

	if not selectors.is_campaign_data_source(campaign=campaign, data_source=data_source):
		return

	for campaign_data_source in models.CampaignDataSources.filter(campaign=campaign, data_source=data_source):
		campaign_data_source.delete_instance()


def update_campaign(supervisor: models.Supervisor, name: str, start_ts: dt, end_ts: dt, data_sources: List[models.DataSource]) -> None:
	"""
	Update parameters of a campaign object in the database.
	:param supervisor: supervisor of the campaign (includes reference to user and campaign)
	:param name: title of the campaign
	:param start_ts: when campaign starts
	:param end_ts: when campaign ends
	:param data_sources: list of data sources
	:return: newly created Campaign instance
	"""

	campaign: models.Campaign = notnull(supervisor).campaign
	campaign.name = notnull(name)
	campaign.start_ts = notnull(start_ts)
	campaign.end_ts = notnull(end_ts)
	campaign.save()

	olds = set(selectors.get_campaign_data_sources(campaign=campaign))
	news = set(data_sources)

	for old in olds.difference(news):
		remove_campaign_data_source(campaign=campaign, data_source=old)

	for new in news.difference(olds):
		add_campaign_data_source(campaign=campaign, data_source=new)


def delete_campaign(supervisor: models.Supervisor) -> None:
	"""
	Delete a campaign - must only be called if campaign's owner makes the call.
	:param supervisor: supervisor of the campaign (includes reference to user and campaign)
	:return: None
	"""

	campaign: models.Campaign = notnull(supervisor).campaign
	if supervisor.user == campaign.owner:
		campaign.delete_instance()


def create_data_source(name: str, icon_name: str) -> models.DataSource:
	"""
	Creates a data source (if not exists)
	:param name: name of the data source
	:param icon_name: icon of the data source
	:return: newly created data source (or the one with matching name)
	"""

	data_source = models.DataSource.get_or_none(name=notnull(name))
	if data_source:
		return data_source

	return models.DataSource.create(name=name, icon_name=notnull(icon_name))


def create_data_record(participant: models.Participant, data_source: models.DataSource, ts: dt, val: Dict) -> None:
	"""
	Creates a data record in raw data table (e.g. sensor reading)
	:param participant: participant of a campaign
	:param data_source: data source of the data record
	:param ts: timestamp
	:param val: value
	:return: None
	"""

	wrappers.DataTable.insert(participant=participant, data_source=data_source, ts=ts, val=val)


def create_data_records(participant: models.Participant, data_source_ids: List[int], tss: List[dt], vals: List[Dict]) -> None:
	"""
	Creates a data record in raw data table (e.g. sensor reading)
	:param participant: participant of a campaign
	:param data_source_ids: data sources of the data records
	:param tss: timestamps
	:param vals: values
	:return: None
	"""

	data_sources: Dict[int, models.DataSource] = dict()
	for ts, data_source_id, val in zip(tss, data_source_ids, vals):
		if data_source_id not in data_sources:
			db_data_source = selectors.find_data_source(data_source_id=data_source_id, name=None)
			if db_data_source is None:
				continue
			data_sources[data_source_id] = db_data_source
		create_data_record(participant=participant, data_source=data_sources[data_source_id], ts=ts, val=val)


def dump_data(participant: models.Participant, data_source: Optional[models.DataSource]) -> str:
	"""
	Dumps content of a particular DataTable into a downloadable file
	:param participant: participant that has reference to user and campaign
	:param data_source: which data source to dump
	:return: path to the downloadable file
	"""

	return wrappers.DataTable.dump_to_file(participant=notnull(participant), data_source=data_source)



from os import getenv

# libs
import psycopg2
from psycopg2 import extras
from peewee import PostgresqlDatabase

# app
from .wrappers import DataTable
from .utils import notnull

from . import models, selectors, services, utils, wrappers


def init():
	models.db = PostgresqlDatabase(
		host=notnull(getenv('POSTGRES_HOST')),
		port=notnull(getenv('POSTGRES_PORT')),
		database=notnull(getenv('POSTGRES_DBNAME')),
		user=notnull(getenv('POSTGRES_USER')),
		password=notnull(getenv('POSTGRES_PASSWORD'))
	)

	# create schema if necessary
	conn = psycopg2.connect(
		host=notnull(getenv('POSTGRES_HOST')),
		port=notnull(getenv('POSTGRES_PORT')),
		dbname=notnull(getenv('POSTGRES_DBNAME')),
		user=notnull(getenv('POSTGRES_USER')),
		password=notnull(getenv('POSTGRES_PASSWORD'))
	)
	cur: extras.DictCursor = conn.cursor(cursor_factory=extras.DictCursor)
	cur.execute('create schema if not exists core')
	cur.execute('create schema if not exists data')
	cur.execute('create schema if not exists geoplan')
	conn.commit()
	conn.close()

	# connect and prepare tables
	models.db.connect()
	models.db.create_tables([
		models.User,
		models.Campaign,
		models.DataSource,
		models.CampaignDataSources,
		models.Supervisor,
		models.Participant,
		models.HourlyStats
	])

	DataTable.con = psycopg2.connect(
		database=notnull(getenv('POSTGRES_DBNAME')),
		host=notnull(getenv('POSTGRES_HOST')),
		port=notnull(getenv('POSTGRES_PORT')),
		user=notnull(getenv('POSTGRES_USER')),
		password=notnull(getenv('POSTGRES_PASSWORD')),
		options="-c search_path=data"
	)


__all__ = [models, selectors, services, utils, wrappers, init]



from datetime import datetime as dt
import dateutil.parser as parser
from os.path import join, exists
from typing import Any, List
from os import mkdir, chmod
import tempfile
import hashlib
import re


def replacenull(value: Any, replacement: Any):
	"""
	Validate that a provided argument is not None, replace if so.
	Raises a ValueError if both value and replacement are None.
	:param value: value being checked for None
	:param replacement: replacement if value is None
	:return: value or replacement depending on value's content
	"""

	if value is None:
		if replacement is None:
			raise ValueError(f'replacement value cannot be {None}')
		else:
			return replacement  # is None
	else:
		return value  # is not None


def notnull(value: Any) -> Any:
	"""
	Asserts that a provided argument is not None
	:param value: value being checked
	:return: value if it is not None
	"""

	assert value is not None
	return value


def ts2int(value: dt) -> int:
	return int(round(value.timestamp() * 1000))


def int2ts(value: int) -> dt:
	return dt.fromtimestamp(value / 1000)


def ts2str(ts: dt) -> str:
	if ts is None or ts == 0:
		return "N/A"
	else:
		return ts.strftime('%m/%d (%a), %I:%M %p')


def ts2web(ts: dt) -> str:
	return ts.strftime('%Y-%m-%dT%H:%M')


def parse_ts(s: str) -> dt:
	"""
	Converts a given string to a datetime object
	:param s: timestamp in string format
	:return: datetime object
	"""

	return parser.parse(f'{s} +0900')


def is_numeric(s: str, floating=False) -> bool:
	if floating:
		return re.search(pattern=r'^[+-]?\d+\.\d+$', string=s) is not None
	else:
		return re.search(pattern=r'^[+-]?\d+$', string=s) is not None


def param_check(request_body, params: List[str]) -> bool:
	for param in params:
		if param not in request_body:
			return False
	return True


def now_dt() -> dt:
	return dt.now()


def now_us() -> int:
	return int(now_dt().timestamp() * 1000 * 1000)


def now_ms() -> int:
	return int(now_dt().timestamp() * 1000)


def md5(value: str) -> str:
	return hashlib.md5(value.encode()).hexdigest()


def get_temp_filepath(filename: str) -> str:
	"""
	Validates presence of a temporary directory, and opens a file for writing in the directory.
	:param filename: filename for writing
	:return: path to the file
	"""

	root = join(tempfile.gettempdir(), 'easytrack_dashboard')
	if not exists(root):
		mkdir(root)
		chmod(root, 0o777)

	res = join(root, filename)
	fp = open(res, 'w+', encoding='utf8')
	fp.close()

	return res


def is_web_ts(s: str) -> bool:
	"""
	Checks if the provided string has a valid datetime format
	:param s: string being validated
	:return: whether string is a valid timestsamp
	"""
	return bool(re.search(pattern=r'^\d{4}-\d{1,2}-\d{1,2}T\d{1,2}:\d{1,2}$', string=s))



from datetime import datetime as dt
from datetime import timedelta as td
from typing import Optional

# libs
from peewee import AutoField, TextField, ForeignKeyField, TimestampField
from peewee import Model, PostgresqlDatabase
from playhouse.postgres_ext import BinaryJSONField

db: Optional[PostgresqlDatabase] = None


class User(Model):
	id = AutoField(primary_key=True, null=False)
	email = TextField(unique=True, null=False)
	name = TextField(null=False)
	session_key = TextField(default=None, null=True)
	tag = TextField(default=None, null=True)

	class Meta:
		database = db
		db_table = 'user'
		schema = 'core'


class Campaign(Model):
	id = AutoField(primary_key=True, null=False)
	owner = ForeignKeyField(User, on_delete='CASCADE', null=False)
	name = TextField(null=False)
	start_ts = TimestampField(default=dt.now, null=False)
	end_ts = TimestampField(default=lambda: dt.now() + td(days=90), null=False)

	class Meta:
		database = db
		db_table = 'campaign'
		schema = 'core'


class DataSource(Model):
	id = AutoField(primary_key=True, null=False)
	name = TextField(unique=True, null=False)
	icon_name = TextField(null=False)

	class Meta:
		database = db
		db_table = 'data_source'
		schema = 'core'


class CampaignDataSources(Model):
	campaign = ForeignKeyField(Campaign, on_delete='CASCADE', null=False)
	data_source = ForeignKeyField(DataSource, on_delete='CASCADE', null=False)

	class Meta:
		database = db
		db_table = 'campaign_data_source'
		schema = 'core'
		indexes = (
			(('campaign', 'data_source'), True),  # unique together
		)


class Supervisor(Model):
	campaign = ForeignKeyField(Campaign, on_delete='CASCADE', null=False)
	user = ForeignKeyField(User, on_delete='CASCADE', null=False)

	class Meta:
		database = db
		db_table = 'supervisor'
		schema = 'core'
		indexes = (
			(('campaign', 'user'), True),  # unique together
		)


class Participant(Model):
	campaign = ForeignKeyField(Campaign, on_delete='CASCADE', null=False)
	user = ForeignKeyField(User, on_delete='CASCADE', null=False)
	join_ts = TimestampField(default=dt.now, null=False)
	last_heartbeat_ts = TimestampField(default=dt.now, null=False)

	class Meta:
		database = db
		db_table = 'participant'
		schema = 'core'

		indexes = (
			(('campaign', 'user'), True),  # unique together
		)


class HourlyStats(Model):
	participant = ForeignKeyField(Participant, on_delete='CASCADE', null=False)
	data_source = ForeignKeyField(DataSource, on_delete='CASCADE', null=False)
	ts = TimestampField(null=False)
	amounts = BinaryJSONField()

	class Meta:
		database = db
		db_table = 'hourly_stats'
		schema = 'core'

		indexes = (
			(('participant', 'data_source'), False),  # selection by participant and/or data source
			(('ts',), False),  # selection by timestamp
		)



from datetime import datetime as dt
from datetime import timedelta as td
from typing import Dict, List, Optional, OrderedDict
import collections
from os import getenv

# libs
import psycopg2 as pg2
import psycopg2.extras as pg2_extras
from psycopg2.extras import _connection as PostgresConnection  # noqa

# app
from . import models
from .utils import notnull, get_temp_filepath


class DataRecord:

	def __init__(self, data_source: models.DataSource, ts: dt, val: Dict):
		self.data_source: models.DataSource = notnull(data_source)
		self.ts: dt = notnull(ts)
		self.val: Dict = notnull(val)


class DataTable:
	con: Optional[PostgresConnection] = None

	@staticmethod
	def __get_name(participant: models.Participant) -> str:
		"""
		Returns a table name for particular campaign participant
		:param participant: the participant that includes campaign and user id information
		:return: name of the corresponding data table
		"""

		return '_'.join([f'campaign{participant.campaign.id}', f'user{participant.id}'])

	@staticmethod
	def create(participant: models.Participant) -> None:
		"""
		Creates a table for a participant to store their data
		:param participant: user participating in a campaign
		:return:
		"""

		table_name = DataTable.__get_name(participant=participant)

		cur = DataTable.con.cursor()
		cur.execute('create schema if not exists data')
		cur.execute(f'create table if not exists data.{table_name}(data_source_id int references core.data_source (id), ts timestamp, val jsonb)')  # noqa
		cur.execute(f'create index if not exists idx_{table_name}_ts on data.{table_name} (ts)')  # noqa
		cur.close()
		DataTable.con.commit()

	@staticmethod
	def insert(participant: models.Participant, data_source: models.DataSource, ts: dt, val: Dict) -> None:
		"""
		Creates a data record in raw data table (e.g. sensor reading)
		:param participant: participant of a campaign
		:param data_source: data source of the data record
		:param ts: timestamp
		:param val: value
		:return: None
		"""

		table_name = DataTable.__get_name(participant=participant)

		cur = DataTable.con.cursor()
		cur.execute(
			f'insert into data.{table_name}(data_source_id, ts, val) values (%s,%s,%s)',  # noqa
			(  # noqa
				data_source, ts, pg2_extras.Json(val)))
		cur.close()
		DataTable.con.commit()

	@staticmethod
	def select_next_k(participant: models.Participant, data_source: models.DataSource, from_ts: dt, limit: int) -> List[DataRecord]:
		"""
		Retrieves next k data records from database
		:param participant: participant that has refernece to user and campaign
		:param data_source: type of data to retrieve
		:param from_ts: starting timestamp
		:param limit: max amount of records to query
		:return: list of data records
		"""

		table_name = DataTable.__get_name(participant=participant)

		cur: pg2_extras.DictCursor = DataTable.con.cursor(cursor_factory=pg2_extras.DictCursor)
		cur.execute(
			f'select data_source_id, ts, val from data.{table_name} where data_source_id = %s and ts >= %s limit %s',  # noqa
			(  # noqa
				data_source.id, from_ts, limit))
		rows = cur.fetchall()
		cur.close()

		return list(map(lambda row: DataRecord(data_source=models.DataSource.get_by_id(pk=row['data_source_id']), ts=row['ts'], val=row['val']), rows))

	@staticmethod
	def select_range(participant: models.Participant, data_source: models.DataSource, from_ts: dt, till_ts: dt) -> List[DataRecord]:
		"""
		Retrieves filtered data based on provided range (start and end timestamps)
		:param participant: participant that has refernece to user and campaign
		:param data_source: type of data to retrieve
		:param from_ts: starting timestamp
		:param till_ts: ending timestamp
		:return: list of data records
		"""

		table_name = DataTable.__get_name(participant=participant)

		cur: pg2_extras.DictCursor = DataTable.con.cursor(cursor_factory=pg2_extras.DictCursor)
		cur.execute(
			f'select data_source_id, ts, val from data.{table_name} where data_source_id = %s and ts >= %s and ts < %s',  # noqa
			(  # noqa
				data_source.id, from_ts, till_ts))
		rows = cur.fetchall()
		cur.close()

		return list(map(lambda row: DataRecord(data_source=models.DataSource.get_by_id(pk=row.data_source_id), ts=row.ts, val=row.val), rows))

	@staticmethod
	def dump_to_file(participant: models.Participant, data_source: Optional[models.DataSource]) -> str:
		"""
		Dumps content of a particular DataTable into a downloadable file
		:param participant: participant that has reference to user and campaign
		:param data_source: which data source to dump
		:return: path to the downloadable file
		"""

		table_name = DataTable.__get_name(participant=participant)
		res_filepath = get_temp_filepath(filename=f'{table_name}.csv')

		cur: pg2_extras.DictCursor = DataTable.con.cursor()
		if data_source is None:
			with open(file=res_filepath, mode='w') as file:
				cur.copy_to(file=file, table=table_name, sep=',')
		else:
			with open(file=res_filepath, mode='w') as file:
				cur.copy_expert(file=file, sql=f"copy {table_name} to {res_filepath} with (format csv, delimiter ',', quote '\"')")
		cur.close()

		return res_filepath


class DataSourceStats:

	def __init__(self, data_source: models.DataSource, amount_of_samples: Optional[int] = 0, last_sync_time: Optional[dt] = dt.fromtimestamp(0)):
		self.data_source: models.DataSource = notnull(data_source)
		self.amount_of_samples: int = notnull(amount_of_samples)
		self.last_sync_time: dt = notnull(last_sync_time)


class ParticipantStats:

	def __init__(self, participant: models.Participant):
		self.participant: models.Participant = notnull(participant)

		self.per_data_source_stats: OrderedDict[models.DataSource, DataSourceStats] = collections.OrderedDict()
		self.amount_of_data: int = 0
		self.last_sync_ts: dt = dt.fromtimestamp(0)
		data_sources: List[models.DataSource] = list(map(lambda x: x.data_source, models.CampaignDataSources.filter(campaign=participant.campaign)))
		for data_source in sorted(data_sources, key=lambda x: x.name):
			latest_hourly_stats: models.HourlyStats = models.HourlyStats.filter(
				participant=participant,
				data_source=data_source
			).order_by(models.HourlyStats.ts, models.HourlyStats.ts.desc()).limit(1)

			self.per_data_source_stats[data_source] = DataSourceStats(
				data_source=data_source,
				amount_of_samples=latest_hourly_stats.amounts,
				last_sync_time=latest_hourly_stats.ts
			) if latest_hourly_stats else DataSourceStats(data_source=data_source)

			self.amount_of_data += self.per_data_source_stats[data_source].amount_of_samples
			self.last_sync_ts = max(self.last_sync_ts, self.per_data_source_stats[data_source].last_sync_time)

		then = self.participant.join_ts.replace(hour=0, minute=0, second=0, microsecond=0) + td(days=1)  # noqa
		now = dt.now().replace(hour=0, minute=0, second=0, microsecond=0) + td(days=1)
		self.participation_duration: int = (now - then).days

	def __getitem__(self, data_source: models.DataSource) -> DataSourceStats:
		if data_source in self.per_data_source_stats:
			return self.per_data_source_stats[data_source]
		else:
			return DataSourceStats(data_source=data_source)



from setuptools import setup, find_packages

setup(
	name='python-easytrack',
	version='1.0.5',
	license='MIT',
	author="Kobiljon Toshnazarov",
	author_email='kobiljon.toshnazarov@gmail.com',
	packages=find_packages('src'),
	package_dir={'': 'src'},
	url='https://github.com/easy-track/easytrack',
	keywords='easytrack boilerplate',
	install_requires=['psycopg2-binary', 'peewee', 'python-dateutil'],
)
